/*============================================================================*/
/* File: KaitenkeiTakinoRPM.c                                                 */
/* Compiler: MPLAB_X_IDE & MPLAB_XC8                                          */
/* MPU:PIC18F25K50    48MHz(16MHz * 3PLL)                                     */
/*                                                                            */
/* TMR2:1[msec]period                                                         */
/*                                                                            */
/* SW1:設定モード変更、保存                                                   */
/* SW2:キャンセル                                                             */
/* SW3:桁移動                                                                 */
/*                                                                            */
/* Version 1.00 2015-11-14                                                    */
/* Version 2.00  2016-11-02 large display support                             */
/*============================================================================*/

#include <p18f25k50.h>
#include <pic18.h>

#include <stdio.h>
#include <stdlib.h>

#define VERSION_STRING "     200"

#define _XTAL_FREQ 48000000

#pragma config PLLSEL   = PLL3X     // PLL Selection (3x clock multiplier)
#pragma config CFGPLLEN = ON        // PLL Enable Configuration bit (PLL Disabled (firmware controlled))
#pragma config CPUDIV   = NOCLKDIV  // CPU System Clock Postscaler (CPU uses system clock (no divide))
#pragma config LS48MHZ  = SYS48X8   // Low Speed USB mode with 48 MHz system clock (System clock at 48 MHz, USB clock divider is set to 8)
#pragma config FOSC     = HSH       // Oscillator Selection (Internal oscillator)
#pragma config PCLKEN   = OFF       // Primary Oscillator Shutdown (Primary oscillator shutdown firmware controlled)
#pragma config FCMEN    = OFF       // Fail-Safe Clock Monitor (Fail-Safe Clock Monitor disabled)
#pragma config IESO     = OFF       // Internal/External Oscillator Switchover (Oscillator Switchover mode disabled)
#pragma config nPWRTEN  = OFF       // Power-up Timer Enable (Power up timer disabled)
#pragma config BOREN    = OFF       // Brown-out Reset Enable (BOR enabled in hardware (SBOREN is ignored))
#pragma config BORV     = 190       // Brown-out Reset Voltage (BOR set to 1.9V nominal)
#pragma config nLPBOR   = ON        // Low-Power Brown-out Reset (Low-Power Brown-out Reset enabled)
#pragma config WDTEN    = OFF       // Watchdog Timer Enable bits (WDT controlled by firmware (SWDTEN enabled))
#pragma config WDTPS    = 32768     // Watchdog Timer Postscaler (1:32768)
#pragma config CCP2MX   = RC1       // CCP2 MUX bit (CCP2 input/output is multiplexed with RC1)
#pragma config PBADEN   = OFF       // PORTB A/D Enable bit (PORTB<5:0> pins are configured as digital I/O on Reset)
#pragma config T3CMX    = RC0       // Timer3 Clock Input MUX bit (T3CKI function is on RC0)
#pragma config SDOMX    = RB3       // SDO Output MUX bit (SDO function is on RB3)
#pragma config MCLRE    = OFF       // Master Clear Reset Pin Enable (MCLR pin enabled; RE3 input disabled)
#pragma config STVREN   = ON        // Stack Full/Underflow Reset (Stack full/underflow will cause Reset)
#pragma config LVP      = OFF       // Single-Supply ICSP Enable bit (Single-Supply ICSP disabled)
#pragma config ICPRT    = OFF       // Dedicated In-Circuit Debug/Programming Port Enable (ICPORT disabled)
#pragma config XINST    = OFF       // Extended Instruction Set Enable bit (Instruction set extension and Indexed Addressing mode disabled)
#pragma config CP0      = OFF       // Block 0 Code Protect (Block 0 is not code-protected)
#pragma config CP1      = OFF       // Block 1 Code Protect (Block 1 is not code-protected)
#pragma config CP2      = OFF       // Block 2 Code Protect (Block 2 is not code-protected)
#pragma config CP3      = OFF       // Block 3 Code Protect (Block 3 is not code-protected)
#pragma config CPB      = OFF       // Boot Block Code Protect (Boot block is not code-protected)
#pragma config CPD      = OFF       // Data EEPROM Code Protect (Data EEPROM is not code-protected)
#pragma config WRT0     = OFF       // Block 0 Write Protect (Block 0 (0800-1FFFh) is not write-protected)
#pragma config WRT1     = OFF       // Block 1 Write Protect (Block 1 (2000-3FFFh) is not write-protected)
#pragma config WRT2     = OFF       // Block 2 Write Protect (Block 2 (04000-5FFFh) is not write-protected)
#pragma config WRT3     = OFF       // Block 3 Write Protect (Block 3 (06000-7FFFh) is not write-protected)
#pragma config WRTC     = OFF       // Configuration Registers Write Protect (Configuration registers (300000-3000FFh) are not write-protected)
#pragma config WRTB     = OFF       // Boot Block Write Protect (Boot block (0000-7FFh) is not write-protected)
#pragma config WRTD     = OFF       // Data EEPROM Write Protect (Data EEPROM is not write-protected)
#pragma config EBTR0    = OFF       // Block 0 Table Read Protect (Block 0 is not protected from table reads executed in other blocks)
#pragma config EBTR1    = OFF       // Block 1 Table Read Protect (Block 1 is not protected from table reads executed in other blocks)
#pragma config EBTR2    = OFF       // Block 2 Table Read Protect (Block 2 is not protected from table reads executed in other blocks)
#pragma config EBTR3    = OFF       // Block 3 Table Read Protect (Block 3 is not protected from table reads executed in other blocks)
#pragma config EBTRB    = OFF       // Boot Block Table Read Protect (Boot block is not protected from table reads executed in other blocks)


//----------- Sensor ----------------------------------------------
#define SENSOR1       PORTBbits.RB2
#define SENSOR1_TRIS  TRISBbits.RB2
#define SENSOR2       PORTBbits.RB7
#define SENSOR2_TRIS  TRISBbits.RB7
#define SENSOR3       PORTBbits.RB5
#define SENSOR3_TRIS  TRISBbits.RB5
#define SENSOR4       PORTBbits.RB6
#define SENSOR4_TRIS  TRISBbits.RB6

#define SENSOR1_ANSEL ANSELBbits.ANSB2
#define SENSOR3_ANSEL ANSELBbits.ANSB5

#define sensor1_interrupt_enable      INTCON3bits.INT2IE
#define sensor1_interrupt_occurred    INTCON3bits.INT2IF
#define sensor1_interrupt_proority    INTCON3bits.INT2IP
#define sensor1_interrupt_rising_edge INTCON2bits.INTEDG2

#define sensor2_interrupt_enable      IOCBbits.IOCB7
#define sensor3_interrupt_enable      IOCBbits.IOCB5
#define sensor4_interrupt_enable      IOCBbits.IOCB6

#define sensor234_interrupt_Port_Change_Priority  INTCON2bits.IOCIP
#define sensor234_interrupt_Port_Change_Enable    INTCONbits.IOCIE
#define sensor234_Interrupt_Port_Change_Occured   INTCONbits.IOCIF

//----------- 3mm Red LED -----------------------------------------
#define LED1       LATBbits.LATB0
#define LED1_TRIS  TRISBbits.RB0
#define LED2       LATAbits.LATA4
#define LED2_TRIS  TRISAbits.RA4
#define LED3       LATAbits.LATA5
#define LED3_TRIS  TRISAbits.RA5
#define LED4       LATCbits.LATC7
#define LED4_TRIS  TRISCbits.RC7
#define LED5       LATCbits.LATC6
#define LED5_TRIS  TRISCbits.RC6

#define LED_ON  1
#define LED_OFF 0

//----------- Push Switch -----------------------------------------
#define SW1        PORTAbits.RA1
#define SW1_TRIS   TRISAbits.RA1
#define SW2        PORTAbits.RA0
#define SW2_TRIS   TRISAbits.RA0
#define SW3        PORTEbits.RE3

#define SW1_ANSEL ANSELAbits.ANSA1  // digital input RA1
#define SW2_ANSEL ANSELAbits.ANSA0  // digital input RA0

#define SW_ON  0
#define SW_OFF 1

#define DEFAULT_ON_TIME   20  //  20[msec]
#define DEFAULT_OFF_TIME  50  //  50[msec]

//----------- Relay -----------------------------------------------
#define RELAY1        LATCbits.LATC0
#define RELAY1_TRIS   TRISCbits.RC0
#define RELAY2        LATCbits.LATC1
#define RELAY2_TRIS   TRISCbits.RC1

#define RELAY_ON  1
#define RELAY_OFF 0

//----------- LED Driver MAX7219 ----------------------------------
#define MAX7219_LOAD      LATBbits.LATB4
#define MAX7219_LOAD_TRIS TRISBbits.RB4
#define MAX7219_DIN       LATBbits.LATB3
#define MAX7219_DIN_TRIS  TRISBbits.RB3
#define MAX7219_CLK       LATBbits.LATB1
#define MAX7219_CLK_TRIS  TRISBbits.RB1

#define MAX7219_LOAD_ANSEL ANSELBbits.ANSB4
#define MAX7219_DIN_ANSEL  ANSELBbits.ANSB3
#define MAX7219_CLK_ANSEL  ANSELBbits.ANSB1

#define Digit_0       0x7E
#define Digit_1       0x30
#define Digit_2       0x6D
#define Digit_3       0x79
#define Digit_4       0x33
#define Digit_5       0x5B
#define Digit_6       0x5F
#define Digit_7       0x70
#define Digit_8       0x7F
#define Digit_9       0x7B
#define Digit_E       0x4F
#define Digit_F       0x47
#define Digit_U       0x3E
#define Digit_u       0x1C
#define Digit_L       0x0E
#define Digit_C       0x4E
#define Digit_P       0x67
#define Digit_Y       0x3B
#define Digit_A       0x77
#define Digit_a       0x7D
#define Digit_r       0x05
#define Digit_n       0x15
#define Digit_d       0x3D
#define Digit_t       0x0F
#define Digit_G       0x5E
#define Digit_b       0x1F
#define Digit_i       0x04
#define Digit_l       0x30
#define Digit_NON     0x00
#define Digit_o       0x1D
#define Digit_S       0x5B
#define Digit_h       0x17
#define Digit_H       0x37
#define Digit_c       0x0D
#define Digit_US      0x08
#define Digit_MINUS   0x01
#define Digit_DP      0x80

#define ROW1        1
#define ROW2        2

//----------- External LARGE DISPLAY -----------------------------
#define LED7SEG_ENABLE  LATCbits.LATC6
#define LED7SEG_CLK     LATBbits.LATB1
#define LED7SEG_DATA    LATBbits.LATB3

#define CMD_DUMMY        0b00000000
#define CMD_NORMAL       0b00010000
#define CMD_NORMAL_DP    0b00100000
#define CMD_ROTATION     0b00110000
#define CMD_ROTATION_DP  0b01000000
#define CMD_CODE         0b01110000
#define CMD_ID           0b10000000

#define CMD_ASCII           0
#define CMD_MAX7219         1
#define CMD_ASCII_ROTATE    2
#define CMD_MAX7219_ROTATE  3

//----------- Rotary Encoder --------------------------------------
#define ENCODER_A       PORTAbits.RA3
#define ENCODER_A_TRIS  TRISAbits.RA3
#define ENCODER_B       PORTAbits.RA2
#define ENCODER_B_TRIS  TRISAbits.RA2

#define ENCODER_A_ANSEL ANSELAbits.ANSA3  // digital input RA3
#define ENCODER_B_ANSEL ANSELAbits.ANSA2  // digital input RA2

#define CW             1
#define CCW            2
#define ENCODER_STOP   0

unsigned char Rotary_Enc_A;
unsigned char Rotary_Enc_B;

//----------- Buzzer ---------------------------------------------
#define BUZZER       LATCbits.LATC2
#define BUZZER_TRIS  TRISCbits.RC2

//----------- PORT Input/Output ----------------------------------
#define PORT_OUTPUT 0
#define PORT_INPUT  1

//-------- TMR0 --------------------------------------------------
#define Timer0_overflowd              INTCONbits.TMR0IF
#define Timer0_interrupt_enable       INTCONbits.TMR0IE
#define Timer0_priority               INTCON2bits.TMR0IP
#define Timer0_enable                 T0CONbits.TMR0ON
#define Timer0_mode_8bit              T0CONbits.T08BIT
#define Timer0_use_external_clock     T0CONbits.T0CS
#define Timer0_prescale_NOT_assign    T0CONbits.PSA
#define Timer0_prescale_bit2          T0CONbits.T0PS2
#define Timer0_prescale_bit1          T0CONbits.T0PS1
#define Timer0_prescale_bit0          T0CONbits.T0PS0

//-------- TMR1 --------------------------------------------------
#define Timer1_overflowd              PIR1bits.TMR1IF
#define Timer1_interrupt_enable       PIE1bits.TMR1IE
#define Timer1_prescale_bit0          T1CONbits.T1CKPS0
#define Timer1_prescale_bit1          T1CONbits.T1CKPS1
#define Timer1_use_external_clock     T1CONbits.TMR1CS
#define Timer1_enable                 T1CONbits.TMR1ON
#define Timer1_16bitmode              T1CONbits.RD16
#define Timer1_GateControlEnable      T1GCONbits.TMR1GE
#define Timer1_priority               IPR1bits.TMR1IP

//-------- TMR2(1ms period) --------------------------------------
#define Timer2_overflowd              PIR1bits.TMR2IF
#define Timer2_interrupt_enable       PIE1bits.TMR2IE
#define Timer2_prescale_bit0          T2CONbits.T2CKPS0
#define Timer2_prescale_bit1          T2CONbits.T2CKPS1
#define Timer2_postscale_bit0         T2CONbits.T2OUTPS0
#define Timer2_postscale_bit1         T2CONbits.T2OUTPS1
#define Timer2_postscale_bit2         T2CONbits.T2OUTPS2
#define Timer2_postscale_bit3         T2CONbits.T2OUTPS3
#define Timer2_period                 PR2
#define Timer2_enable                 T2CONbits.TMR2ON
#define Timer2_priority               IPR1bits.TMR2IP

//-------- Interrupt ---------------------------------------------
#define Interrupt_High_Enable INTCONbits.GIEH
#define Interrupt_Low_Enable  INTCONbits.GIEL
#define Interrupt_Priority_Level_Enable  RCONbits.IPEN

#define HIGH_PRIORITY 1
#define LOW_PRIORITY  0

//-----　型変換用の共用体 ----------------------------------------
typedef union {
  unsigned long  Lval;
  unsigned short Wval[2];
  unsigned char  Bval[4];
} long_union;

typedef union {
  short          Sval;
  unsigned int   Ival;
  unsigned short Wval;
  unsigned char  Bval[2];
} short_union;

//-------- Gloval variable (common function) ---------------------
unsigned int sw1_disable_time;
unsigned int sw2_disable_time;
unsigned int sw3_disable_time;
unsigned int sw1_disable_counter;
unsigned int sw2_disable_counter;
unsigned int sw3_disable_counter;
BOOL         sw1_enable;
BOOL         sw2_enable;
BOOL         sw3_enable;

//-------- Gloval variable (external display) --------------------
char extled_display_buffer[16];  // index 0 = upper right  index 15 = lower left
BOOL extled_display_update[16];

//-------- EEPROM initialize -------------------------------------
__EEPROM_DATA(0x01, 1, 1, 0xF4,0x1,0xB8,0xB, 1); // address=0  1F4 = 500   BB8 = 3000
__EEPROM_DATA(0,0,0,0, 100,0,0,0);                  // address=8  rpm target Lower & Upper
__EEPROM_DATA(1,0,0,0, 0, 0xFF,0xFF,0xFF);          // address=16 ギヤ比 + 小数点位置
__EEPROM_DATA(1,0,0,0, 200,0,0,0);                  // address=24 最少直径　最大直径
__EEPROM_DATA(100,0,0,0, 5,0,0,0);                  // address=32 現在直径　増減値
__EEPROM_DATA(100,0,0,0, 200,0,0,0);                // address=40 speed target Lower & Upper

#define EEPROM_ADDRESS_LED_INTENSITY       0
#define EEPROM_ADDRESS_RPM_RELAY_ENABLE    1
#define EEPROM_ADDRESS_SPEED_ENABLE        2
#define EEPROM_ADDRESS_UPDATE_TIME         3  // msec
#define EEPROM_ADDRESS_ZERO_CLEAR_TIME     5  // msec
#define EEPROM_ADDRESS_SPEED_RELAY_ENABLE  7
#define EEPROM_ADDRESS_RPM_TARGET_LOWER    8  // リレーOFF RPM
#define EEPROM_ADDRESS_RPM_TARGET_UPPER   12  // リレーON  RPM
#define EEPROM_ADDRESS_GEAR_RATIO         16  // ギヤ比
#define EEPROM_ADDRESS_GEAR_RATIO_DP      20  // ギヤ比 小数点位置
#define EEPROM_ADDRESS_DIAMETER_MIN       24  // 最少直径mm
#define EEPROM_ADDRESS_DIAMETER_MAX       28  // 最大直径mm
#define EEPROM_ADDRESS_DIAMETER_CURRENT   32  // 現在直径mm
#define EEPROM_ADDRESS_DIAMETER_STEP      36  // 直径増減値mm
#define EEPROM_ADDRESS_SPEED_TARGET_LOWER 40  // リレーOFFスピード
#define EEPROM_ADDRESS_SPEED_TARGET_UPPER 44  // リレーON スピード

#define ENABLE  1
#define DISABLE 0

//-------- application Gloval variable & constants ---------------
#define CONST_MIN           720000000 // TMR1 count 1 min @ 48MHz
#define ZERO_CLEAR_TIME          3000 // 1500 x 1ms = 3.0 sec
#define TIMER1_OVERFLOW_TIME     5.46 // timer1 overflow time msec

BOOL start_flag;
BOOL display_flag;

unsigned char TMR1L_value;
unsigned char TMR1H_value;
unsigned int  TMR1_OVF_value;
unsigned int  of_count;  // over flow counter
unsigned long rotate;
unsigned long rotate_value;
unsigned int  zero_display_counter;
unsigned int  display_update_count;
unsigned int  display_diameter_counter;

typedef union {
  unsigned long ulong;
  unsigned int  uint[2];
  unsigned char uchar[4];
 //  double dbl;
} Union4bytes;


//------------------------------------------------------------------------------
// High priority interrupt
//------------------------------------------------------------------------------
void interrupt High_Priority_Interrupt(void)
{
  if( sensor1_interrupt_occurred ) {
    if( start_flag ) {   // save TMR1 value
      TMR1L_value = TMR1L;
      TMR1H_value = TMR1H;
      rotate++;
      if( of_count >= display_update_count ) {
        TMR1_OVF_value = of_count;
        TMR1H = 0;
        TMR1L = 0;
        rotate_value = rotate;
        of_count = 0;
        rotate   = 0;
        display_flag = TRUE;
      }
    } else {
      TMR1H = 0;
      TMR1L = 0;
      of_count = 0;
      rotate   = 0;
      start_flag = TRUE;
    }
    zero_display_counter = 0;
    sensor1_interrupt_occurred = FALSE;
  }
}

//------------------------------------------------------------------------------
// Low priority interrupt
//------------------------------------------------------------------------------
void interrupt low_priority Low_Priority_Interrupt(void)
{
  if( Timer1_overflowd ) {
    Timer1_overflowd = FALSE;
    of_count++;
  }

  if( Timer2_overflowd ) {   // 1msec period
    //------------- SW1 ------------------
    if( SW1 == SW_ON ) {
      sw1_disable_counter = sw1_disable_time;
    } else {
      if( sw1_disable_counter > 0 ) {
        sw1_disable_counter--;
      } else {
        sw1_enable = TRUE;
      }
    }
    //------------- SW2 ------------------
    if( SW2 == SW_ON ) {
      sw2_disable_counter = sw2_disable_time;
    } else {
      if( sw2_disable_counter > 0 ) {
        sw2_disable_counter--;
      } else {
        sw2_enable = TRUE;
      }
    }
    //------------- SW3 ------------------
    if( SW3 == SW_ON ) {
      sw3_disable_counter = sw3_disable_time;
    } else {
      if( sw3_disable_counter > 0 ) {
        sw3_disable_counter--;
      } else {
        sw3_enable = TRUE;
      }
    }

    zero_display_counter++;
    if( 0 < display_diameter_counter ) {
      display_diameter_counter--;
    }
    Timer2_overflowd = FALSE;
  }
}

//----------------------------------------------------------------------------
// ２バイトをＥＥＰＲＯＭから読み出し
//----------------------------------------------------------------------------
unsigned int read_word_eeprom( unsigned int addr )
{
  short_union udata;

  udata.Bval[0] = eeprom_read( addr + 0 );
  udata.Bval[1] = eeprom_read( addr + 1 );

  return udata.Ival;
}

//----------------------------------------------------------------------------
// ２バイトをＥＥＰＲＯＭへ書き込み
//----------------------------------------------------------------------------
void write_word_eeprom( unsigned int addr, unsigned int value )
{
  short_union udata;

  udata.Ival = value;
  eeprom_write( addr+0, udata.Bval[0] );
  eeprom_write( addr+1, udata.Bval[1] );
}

//----------------------------------------------------------------------------
// ４バイトをＥＥＰＲＯＭから読み出し
//----------------------------------------------------------------------------
unsigned long read_long_eeprom( unsigned int addr )
{
  long_union udata;

  udata.Bval[0] = eeprom_read( addr + 0 );
  udata.Bval[1] = eeprom_read( addr + 1 );
  udata.Bval[2] = eeprom_read( addr + 2 );
  udata.Bval[3] = eeprom_read( addr + 3 );

  return udata.Lval;
}

//----------------------------------------------------------------------------
// ４バイトをＥＥＰＲＯＭへ書き込み
//----------------------------------------------------------------------------
void write_long_eeprom( unsigned int addr, unsigned long value )
{
  long_union udata;

  udata.Lval = value;
  eeprom_write( addr + 0, udata.Bval[0] );
  eeprom_write( addr + 1, udata.Bval[1] );
  eeprom_write( addr + 2, udata.Bval[2] );
  eeprom_write( addr + 3, udata.Bval[3] );
}

/*
//----------------------------------------------------------------------------
// ４バイトをＥＥＰＲＯＭから読み出し
//----------------------------------------------------------------------------
double read_double_eeprom( unsigned int addr )
{
  Union4bytes udata;

  udata.uchar[0] = eeprom_read( addr + 0 );
  udata.uchar[1] = eeprom_read( addr + 1 );
  udata.uchar[2] = eeprom_read( addr + 2 );
  udata.uchar[3] = eeprom_read( addr + 3 );

  return udata.dbl;
}

//----------------------------------------------------------------------------
// ４バイトをＥＥＰＲＯＭへ書き込み
//----------------------------------------------------------------------------
void write_double_eeprom( unsigned int addr, double value )
{
  Union4bytes udata;

  udata.dbl = value;
  eeprom_write( addr + 0, udata.uchar[0] );
  eeprom_write( addr + 1, udata.uchar[1] );
  eeprom_write( addr + 2, udata.uchar[2] );
  eeprom_write( addr + 3, udata.uchar[3] );
}
*/

//------------------------------------------------------------------------------
// delay  msec        Clock = 48MHz
//------------------------------------------------------------------------------
void delay_ms(unsigned short delay)
{
  unsigned short i;

  for( i = 0 ; i <= delay ; i++ ) {
    Delay1KTCYx(12);
  }
}

//------------------------------------------------------------------------------
// delay  msec        Clock = 48MHz
//------------------------------------------------------------------------------
void delay_ms2(unsigned short delay)
{
  unsigned short i;

  for( i = 0 ; i <= delay ; i++ ) {
    __delay_ms(1);
  }
}

//-------------------------------------------------------------------------------
// SPI initializes
//-------------------------------------------------------------------------------
void Spi_initialize()
{
  SSP1STATbits.SMP   = 0; // Input data sampled at middle of data output time
  SSP1STATbits.CKE   = 1;
  SSP1CON1bits.CKP   = 0;
  SSP1CON1bits.SSPM3 = 0; // SPI Master mode Clock FSOC/4
  SSP1CON1bits.SSPM2 = 0; // SPI Master mode Clock FSOC/4
  SSP1CON1bits.SSPM1 = 1; // SPI Master mode Clock FSOC/4
  SSP1CON1bits.SSPM0 = 0; // SPI Master mode Clock FSOC/4
  PIR1bits.SSPIF = 0;
  SSP1CON1bits.SSPEN = 1; // Enables serial port
}

//-------------------------------------------------------------------------------
// SPI write
//-------------------------------------------------------------------------------
void Spi_write_data(unsigned char data )
{
  SSP1BUF = data;

  while( PIR1bits.SSPIF == 0 ) {}
  PIR1bits.SSPIF = 0;
}

//-------------------------------------------------------------------------------
// SPI disable
//-------------------------------------------------------------------------------
void Spi_disable()
{
  MAX7219_LOAD = 1;
  SSP1CON1bits.SSPEN = 0; // Disables serial port
}

//-------------------------------------------------------------------------------
// SPI enable
//-------------------------------------------------------------------------------
void Spi_enable()
{
  SSP1CON1bits.SSPEN = 1; // Disables serial port
}

//------------------------------------------------------------------------------
// external LARGE LED display
//------------------------------------------------------------------------------
void led7seg_write( unsigned char data )
{
  LED7SEG_CLK = 1;
  __delay_us(1);
  LED7SEG_ENABLE = 0;
  for( unsigned char i = 0 ; i < 8 ; i++ ) {
    LED7SEG_CLK = 0;
    if( (data & 0x80) == 0x80 ) {
      LED7SEG_DATA = 1;
    } else {
      LED7SEG_DATA = 0;
    }
    data <<= 1;
    __delay_us(10);  //' 25
    LED7SEG_CLK = 1;
    __delay_us(10);
  }
  LED7SEG_ENABLE = 1;
  __delay_us(15);
}

//------------------------------------------------------------------------------
// external LARGE LED display
//------------------------------------------------------------------------------
void initialize_large_display()
{
  Spi_disable();
  for( unsigned char display_id = 0 ; display_id < 16 ; display_id++ ) {
    led7seg_write( CMD_ID | CMD_CODE | display_id );
    led7seg_write( CMD_MAX7219 );
    delay_ms(10);
  }
  Spi_enable();
}

//------------------------------------------------------------------------------
// external LARGE LED display
//------------------------------------------------------------------------------
void update_large_display()
{
  static unsigned char display_id = 0;
  char tmp_data;

  tmp_data = extled_display_buffer[display_id];

  if( extled_display_update[display_id] ) {
    Spi_disable();
    if( tmp_data >= 0x80 ) {
      led7seg_write( CMD_ID | CMD_NORMAL_DP | display_id );
      tmp_data &= 0x7F;
    } else {
      led7seg_write( CMD_ID | CMD_NORMAL | display_id );
    }
    led7seg_write( tmp_data );
    extled_display_update[display_id] = FALSE;

    Spi_enable();
  }

  if( display_id == 15 ) {
    display_id = 0;
  } else {
    display_id++;
  }
  __delay_us(500);
}

//------------------------------------------------------------------------------
// external LARGE LED display
//------------------------------------------------------------------------------
void refresh_large_display()
{
  for( unsigned char i = 0 ; i < 16 ; i++ ) {
    update_large_display();
  }
}

//------------------------------------------------------------------------------
// MAX7219 write data
//------------------------------------------------------------------------------
void max7219write( unsigned char addrmain, unsigned char datamain, unsigned char addrsub, unsigned char datasub )
{
  MAX7219_LOAD = 0;
  Spi_write_data( addrsub );
  Spi_write_data( datasub );
  Spi_write_data( addrmain );
  Spi_write_data( datamain );
  MAX7219_LOAD = 1;
}

//------------------------------------------------------------------------------
// MAX7219 H/W initialize
//------------------------------------------------------------------------------
void max7219_hw_initialize()
{
  MAX7219_LOAD_TRIS  = PORT_OUTPUT;
  MAX7219_DIN_TRIS   = PORT_OUTPUT;
  MAX7219_CLK_TRIS   = PORT_OUTPUT;
  MAX7219_LOAD_ANSEL = 0;
  MAX7219_DIN_ANSEL  = 0;
  MAX7219_CLK_ANSEL  = 0;
}

//------------------------------------------------------------------------------
// MAX7219 initialize   no decode mode
//------------------------------------------------------------------------------
void max7219_initialize()
{
  max7219_hw_initialize();

  delay_ms(100);  // slow start power supply
  max7219write( 0x0F, 0x00, 0x0F, 0x00 ); // display test , normal operation
  max7219write( 0x09, 0x00, 0x09, 0x00 ); // decode mode , 0xFF:decode digit 0 to 7  0x00:NO decode
  max7219write( 0x0A, 0x01, 0x0A, 0x01 ); // intensity , 0x01=dark .... 0x0F=bright
  max7219write( 0x0B, 0x07, 0x0B, 0x07 ); // scan limit , main -> display 0 to 7   sub -> 0 to 2
  max7219write( 0x0C, 0x01, 0x0C, 0x01 ); // shutdown mode register , normal operation
}

//------------------------------------------------------------------------------
// MAX7219 dimmer setup
//
// dimmer: 0x01(dark) 0x0F(bright)
//------------------------------------------------------------------------------
void max7219_intensity(unsigned char dimmer)
{
  max7219write( 0x0A, dimmer, 0x0A, dimmer ); // intensity , 0x01=dark .... 0x0F=bright
}

//------------------------------------------------------------------------------
// MAX7219 code convert
//
// position
//  1:upper left    8:upper right
//  9:lower left   16:lower right
//------------------------------------------------------------------------------
unsigned char max7219_ascii_to_code(unsigned char ascii)
{
  unsigned char char_code;

  switch( ascii ) {
    case '0': char_code = Digit_0;     break;
    case '1': char_code = Digit_1;     break;
    case '2': char_code = Digit_2;     break;
    case '3': char_code = Digit_3;     break;
    case '4': char_code = Digit_4;     break;
    case '5': char_code = Digit_5;     break;
    case '6': char_code = Digit_6;     break;
    case '7': char_code = Digit_7;     break;
    case '8': char_code = Digit_8;     break;
    case '9': char_code = Digit_9;     break;
    case 'E': char_code = Digit_E;     break;
    case 'F': char_code = Digit_F;     break;
    case 'U': char_code = Digit_U;     break;
    case 'u': char_code = Digit_u;     break;
    case 'L': char_code = Digit_L;     break;
    case 'C': char_code = Digit_C;     break;
    case 'P': char_code = Digit_P;     break;
    case 'Y': char_code = Digit_Y;     break;
    case 'A': char_code = Digit_A;     break;
    case 'a': char_code = Digit_a;     break;
    case 'r': char_code = Digit_r;     break;
    case 'n': char_code = Digit_n;     break;
    case 'd': char_code = Digit_d;     break;
    case 't': char_code = Digit_t;     break;
    case 'G': char_code = Digit_G;     break;
    case 'b': char_code = Digit_b;     break;
    case 'i': char_code = Digit_i;     break;
    case 'I': char_code = Digit_l;     break;
    case ' ': char_code = Digit_NON;   break;
    case 'o': char_code = Digit_o;     break;
    case 'S': char_code = Digit_S;     break;
    case 'h': char_code = Digit_h;     break;
    case 'H': char_code = Digit_H;     break;
    case 'c': char_code = Digit_c;     break;
    case '_': char_code = Digit_US;    break;
    case '-': char_code = Digit_MINUS; break;
    case '.': char_code = Digit_DP;    break;
    default:  char_code = 0;           break;
  }

  return char_code;
}

//------------------------------------------------------------------------------
// MAX7219 write string
//
// position
//  1:upper left    8:upper right
//  9:lower left   16:lower right
//------------------------------------------------------------------------------
void max7219_display_char(unsigned char ascii, unsigned char position)
{
  unsigned char char_code;

  char_code =  max7219_ascii_to_code( ascii );
  if( position <= 4 ) {
    max7219write( position , char_code, 0, 0 );
  } else if( position <= 7 ) {
    max7219write( 0, 0, position - 4, char_code );
  } else if( position <= 12 ) {
    max7219write( position - 4, char_code, 0, 0 );
  } else {
    max7219write( 0, 0, position - 8, char_code );
  }
}

//------------------------------------------------------------------------------
// MAX7219 write string
//
// buf:
//   ascii char string
// start:
//   1(left) 8(right)
// len:
//   string length
// row_number:
//   1(upper) 2(lower)
//------------------------------------------------------------------------------
void max7219_write_string( unsigned char buf[], unsigned char start, unsigned char len, unsigned char row_number, unsigned char dp_pos )
{
  unsigned char row_offset;
  unsigned char address[8];
  unsigned char data[8];
  unsigned char start_addr;
  unsigned char buf_index;
  unsigned char tmp_len;
  unsigned char tmp_sorc;
  unsigned char ext_index;

  for( unsigned char index = 0 ; index < 8 ; index++ ) {
    address[index] = 0;
    data[index]    = 0;
  }

  ext_index = (8 - start) + ((row_number - 1) * 8);
  tmp_len = len;
  buf_index = 0;
  while( tmp_len != 0 ) {
    tmp_sorc =  buf[buf_index];
    if( tmp_sorc != extled_display_buffer[ext_index] ) {
      extled_display_buffer[ext_index] = max7219_ascii_to_code(tmp_sorc);
      extled_display_update[ext_index] = TRUE;
    }
    buf_index++;
    ext_index--;
    tmp_len--;
  }
  refresh_large_display();

  row_offset = (row_number - 1) * 4;

  start_addr = start;
  buf_index = 0;
  while( len > 0 ) {
    len--;
    if( start_addr == 5 ) {
      start_addr -= 4;
    }
    address[start-1] = start_addr + row_offset;
    data[start-1]    = max7219_ascii_to_code( buf[buf_index] );
    start++;
    start_addr++;
    buf_index++;
  }

  if( dp_pos != 0 ) {
    data[dp_pos-1] += Digit_DP;
  }
  max7219write( address[0], data[0], address[4], data[4]  );
  max7219write( address[1], data[1], address[5], data[5]  );
  max7219write( address[2], data[2], address[6], data[6]  );
  max7219write( address[3], data[3], address[7], data[7]  );
}

//------------------------------------------------------------------------------
// clear display
//------------------------------------------------------------------------------
void max7219_clear()
{
  for( unsigned char index = 1; index <= 8 ; index++ ) {
    max7219write( index, Digit_NON , index, Digit_NON );
  }
}

//------------------------------------------------------------------------------
// max7219 locate and write
//
// start:       8(left digit) to 1(right digit)
// row_number:  1(upper line) 2(lower line)
//------------------------------------------------------------------------------
void max7219_write_string_number( unsigned char buf[], unsigned char start, unsigned char len, unsigned char row_number )
{
  unsigned char row_offset;
  unsigned char address[8];
  unsigned char data[8];
  unsigned char start_addr;
  unsigned char ext_index;
  unsigned char tmp_len;
  unsigned char buf_index;
  unsigned char tmp_sorc;

  for( unsigned char index = 0 ; index < 8 ; index++ ) {
    address[index] = 0;
    data[index]    = 0;
  }

  ext_index = (start - len) + ((row_number - 1) * 8);
  tmp_len = len;
  buf_index = 0;
  while( tmp_len != 0 ) {
    tmp_sorc =  buf[buf_index];
    if( tmp_sorc != extled_display_buffer[ext_index] ) {
      extled_display_buffer[ext_index] = tmp_sorc;
      extled_display_update[ext_index] = TRUE;
    }
    buf_index++;
    ext_index++;
    tmp_len--;
  }

  row_offset = (row_number - 1) * 4;

  start_addr = 8 - start + 1;
  if( start_addr >= 5 ) {
    start_addr -= 4;
  }
  while( len > 0 ) {
    len--;
    if( start_addr == 5 ) {
      start_addr -= 4;
    }
    address[start-1] = start_addr + row_offset;
    data[start-1]    = buf[len];
    start--;
    start_addr++;
  }

  max7219write( address[7], data[7], address[3], data[3]  );
  max7219write( address[6], data[6], address[2], data[2]  );
  max7219write( address[5], data[5], address[1], data[1]  );
  max7219write( address[4], data[4], address[0], data[0]  );
}

//------------------------------------------------------------------------------
// display number
//
// dp_pos:      0(no display)  1(least significant digit)
// row_number:  1(upper line)  2(lower line)
//------------------------------------------------------------------------------
void led_display_number( long number, unsigned char dp_pos, unsigned char row_number, unsigned char max_digit )
{
  unsigned char dig;
  unsigned char buf[8];
  unsigned long ul_number;
  BOOL          is_negative;

  if( number < 0) {
    ul_number = -number;
    is_negative = TRUE;
  } else {
    ul_number = number;
    is_negative = FALSE;
  }

  for( unsigned char index = 0; index < max_digit ; index++ ) {
    dig = ul_number % 10;
    switch( dig ) {
      case 0:
          if( ul_number == 0 ) {
            if( is_negative ) {
              buf[index] = Digit_MINUS;
              is_negative = FALSE;
            } else {
              if( index == 0 ) {
                buf[index] = Digit_0;
              } else {
                buf[index] = Digit_NON;
              }
            }
          } else {
            buf[index] = Digit_0;
          }
          break;
      case 1: buf[index] = Digit_1; break;
      case 2: buf[index] = Digit_2; break;
      case 3: buf[index] = Digit_3; break;
      case 4: buf[index] = Digit_4; break;
      case 5: buf[index] = Digit_5; break;
      case 6: buf[index] = Digit_6; break;
      case 7: buf[index] = Digit_7; break;
      case 8: buf[index] = Digit_8; break;
      case 9: buf[index] = Digit_9; break;
    }
    if( (dp_pos - 1) == index ) {
       buf[index] += Digit_DP;
    } 
    ul_number /= 10;
  }  

  if( row_number == 1 ) {
    max7219_write_string_number( buf, max_digit, max_digit, ROW1 );
  } else {
    max7219_write_string_number( buf, max_digit, max_digit, ROW2 );
  }
}

//------------------------------------------------------------------------------
// Rotary encoder
//
// return
//    CW / CCW / OTHER
//------------------------------------------------------------------------------
char rotary_encoder()
{
  static BOOL Rotary_A_Lock;
  static BOOL Rotary_B_Lock;
  static char Rotary_Stat;

  unsigned char temp_A;
  unsigned char temp_B;
  char return_value;

  temp_A = ENCODER_A;
  temp_B = ENCODER_B;

  return_value = ENCODER_STOP;

  if( (Rotary_Enc_A == 1) && (temp_A == 0) && (temp_B == 1) ) {
    Rotary_Stat   = CW;
    Rotary_A_Lock = TRUE;
  }
  if( Rotary_A_Lock == TRUE ) {
    if( (temp_A == 0) && (temp_B == 0)) {
      return_value = Rotary_Stat;
      Rotary_A_Lock = FALSE;
    }
  }

  if( (Rotary_Enc_B == 1) && (temp_B == 0) && (temp_A == 1)) {
    Rotary_Stat   = CCW;
    Rotary_B_Lock = TRUE;
  }
  if( Rotary_B_Lock == TRUE ) {
    if( (temp_A == 0) && (temp_B == 0) ) {
      return_value = Rotary_Stat;
      Rotary_B_Lock = FALSE;
    }
  }

  Rotary_Enc_A = temp_A;
  Rotary_Enc_B = temp_B;

  return return_value;
}

//------------------------------------------------------------------------------
// Rotary encoder H/W initialize
//------------------------------------------------------------------------------
void rotary_encoder_hw_initialize()
{
  ENCODER_A_TRIS = PORT_INPUT;
  ENCODER_B_TRIS = PORT_INPUT;

  ENCODER_A_ANSEL = 0;
  ENCODER_B_ANSEL = 0;

  Rotary_Enc_A = ENCODER_A;
  Rotary_Enc_B = ENCODER_B;
}

//------------------------------------------------------------------------------
// Push Switch H/W initialize
//------------------------------------------------------------------------------
void push_switch_hw_initialize()
{
  SW1_TRIS = PORT_INPUT;
  SW2_TRIS = PORT_INPUT;

  SW1_ANSEL = 0;
  SW2_ANSEL = 0;

  sw1_enable = TRUE;
  sw2_enable = TRUE;
  sw3_enable = TRUE;
  sw1_disable_counter = 0;
  sw2_disable_counter = 0;
  sw3_disable_counter = 0;
}

//------------------------------------------------------------------------------
// Relay H/W initialize
//------------------------------------------------------------------------------
void relay_hw_initialize()
{
  RELAY1 = RELAY_OFF;
  RELAY2 = RELAY_OFF;

  RELAY1_TRIS = PORT_OUTPUT;
  RELAY2_TRIS = PORT_OUTPUT;
}

//------------------------------------------------------------------------------
// LED1 - LED5 H/W initialize
//------------------------------------------------------------------------------
void led_hw_initialize()
{
  LED1 = LED_OFF;
  LED2 = LED_OFF;
  LED3 = LED_OFF;
  LED4 = LED_OFF;
  LED5 = LED_OFF;

  LED1_TRIS = PORT_OUTPUT;
  LED2_TRIS = PORT_OUTPUT;
  LED3_TRIS = PORT_OUTPUT;
  LED4_TRIS = PORT_OUTPUT;
  LED5_TRIS = PORT_OUTPUT;
}

//------------------------------------------------------------------------------
// Sensor H/W initialize
//------------------------------------------------------------------------------
void sensor_hw_initialize()
{
  SENSOR1_TRIS = PORT_INPUT;
  SENSOR2_TRIS = PORT_INPUT;
  SENSOR3_TRIS = PORT_INPUT;
  SENSOR4_TRIS = PORT_INPUT;

  SENSOR1_ANSEL = 0;
  SENSOR3_ANSEL = 0;

  sensor1_interrupt_enable = FALSE;
  sensor2_interrupt_enable = FALSE;
  sensor3_interrupt_enable = FALSE;
  sensor4_interrupt_enable = FALSE;
}

//------------------------------------------------------------------------------
// Sensor1 interrupt enable
//------------------------------------------------------------------------------
void sensor1_interupt_enable()
{
  sensor1_interrupt_proority    = HIGH_PRIORITY;
  sensor1_interrupt_rising_edge = FALSE;
  sensor1_interrupt_enable      = TRUE;
  sensor1_interrupt_occurred    = FALSE;
}

//------------------------------------------------------------------------------
// Sensor2,3,4 interrupt enable
//------------------------------------------------------------------------------
void sensor_interupt_enable(unsigned char sensor_number)
{
  unsigned char dummy;

  delay_ms(500);  // slow start sensor
  sensor234_interrupt_Port_Change_Priority = HIGH_PRIORITY;
  sensor234_interrupt_Port_Change_Enable   = TRUE;

  switch( sensor_number ) {
    case 2: sensor2_interrupt_enable = TRUE; dummy = SENSOR2; break;
    case 3: sensor3_interrupt_enable = TRUE; dummy = SENSOR3; break;
    case 4: sensor4_interrupt_enable = TRUE; dummy = SENSOR4; break;
  }

  sensor234_Interrupt_Port_Change_Occured  = FALSE;
}

//------------------------------------------------------------------------------
// TMR1 initialize  free run
//------------------------------------------------------------------------------
void tmr1_initialize()
{
  Timer1_priority  = LOW_PRIORITY;
  Timer1_use_external_clock = FALSE;
  Timer1_prescale_bit0 = 0;  // priscaler 1:1
  Timer1_prescale_bit1 = 0;
  Timer1_overflowd = FALSE;
  Timer1_enable    = TRUE;
  Timer1_GateControlEnable = FALSE;
  Timer1_interrupt_enable = TRUE;
}

//------------------------------------------------------------------------------
// TMR2 initialize  1[ms] period and Interrupt enable
//------------------------------------------------------------------------------
void tmr2_initialize()
{
  Timer2_prescale_bit0 = 0;  // priscaler 1:16
  Timer2_prescale_bit1 = 1;
  Timer2_postscale_bit0 = 1; // postscaler 1:10
  Timer2_postscale_bit1 = 0;
  Timer2_postscale_bit2 = 0;
  Timer2_postscale_bit3 = 1;
  Timer2_period = 74;
  Timer2_priority  = LOW_PRIORITY;
  Timer2_overflowd = FALSE;
  Timer2_enable    = TRUE;
  Timer2_interrupt_enable = TRUE;
}

//------------------------------------------------------------------------------
// push switch ON test
//------------------------------------------------------------------------------
BOOL sw_on( unsigned char sw_no, unsigned int on_time, unsigned int off_time )
{
  unsigned int sw1_on_counter;
  unsigned int sw2_on_counter;
  unsigned int sw3_on_counter;

  sw1_on_counter = 0;
  sw2_on_counter = 0;
  sw3_on_counter = 0;
  switch( sw_no ) {
    case 1:
      while( SW1 == SW_ON && sw1_enable ) {
        delay_ms(1);
        if( sw1_on_counter > on_time ) {
          sw1_disable_time = off_time;
          sw1_enable = FALSE;
          return TRUE;
        }
        sw1_on_counter++;
      }
      break;
    case 2:
      while( SW2 == SW_ON && sw2_enable ) {
        delay_ms(1);
        if( sw2_on_counter > on_time ) {
          sw2_disable_time = off_time;
          sw2_enable = FALSE;
          return TRUE;
        }
        sw2_on_counter++;
      }
      break;
    case 3:
      while( SW3 == SW_ON && sw3_enable ) {
        delay_ms(1);
        if( sw3_on_counter > on_time ) {
          sw3_disable_time = off_time;
          sw3_enable = FALSE;
          return TRUE;
        }
        sw3_on_counter++;
      }
      break;
  }

  return FALSE;
}

//------------------------------------------------------------------------------
// buzzer H/W initialize
//------------------------------------------------------------------------------
void buzzer_hw_initialize()
{
  BUZZER_TRIS = PORT_OUTPUT;
}

//------------------------------------------------------------------------------
// buzzer
//
//  time:beep time[msec]
//------------------------------------------------------------------------------
void beep( unsigned int time )
{
    unsigned int i;

    for(  i = 0 ; i < time ; i++ ) {
      BUZZER = 1;
      __delay_us(500);
      BUZZER = 0;
      __delay_us(500);
    }
}

//----------------------------------------------------------------------------
// unsigned long べき乗
//----------------------------------------------------------------------------
unsigned long pow_long( unsigned long val, unsigned char pv )
{
  unsigned long return_val;

  if( pv == 0 ) {
    return 1;
  }

  return_val = 1;
  for( unsigned char i = 0 ; i < pv ; i++ ) {
    return_val *= val;
  }

  return return_val;
}

//----------------------------------------------------------------------------
// set eeprom default data
//----------------------------------------------------------------------------
void eeprom_set_default()
{
 //   write_long_eeprom(EEPROM_ADDRESS_TARGET_COUNT,0);
    eeprom_write(EEPROM_ADDRESS_LED_INTENSITY,1);
}

//----------------------------------------------------------------------------
// ロータリーエンコーダによる数値入力 long型
//
// init_val:初期値
// min_val :入力可能な最小値
// max_val :入力可能な最大値
//----------------------------------------------------------------------------
unsigned long input_number(
 unsigned long init_val,
 unsigned long min_val,
 unsigned long max_val)
{
  unsigned char turn;     // CW,CCW　ロータリーエンコーダ回転方向
  unsigned char cursor;   // カーソル位置
  unsigned long value;    // 入力した値を格納する変数
  unsigned char digits;   // 桁数
  unsigned long tmp_long; // テンポラリ変数

  tmp_long = max_val;
  digits = 0;
  while( tmp_long > 0 ) {  // 最大値の桁数計算
    tmp_long /= 10;
    digits++;
  }

  cursor = 0;
  value = init_val;
  led_display_number( value, cursor+1, ROW2, digits );
  while( TRUE ) {
    turn = rotary_encoder();
    if( turn != ENCODER_STOP ) {
      if( turn == CW ) {
        if( value + pow_long(10,cursor) > max_val ) {
          value = max_val;
        } else {
          value += pow_long(10,cursor);
          beep(10);
        }
      } else {
        if( value < ( pow_long(10,cursor) + min_val ) ) {
          value = min_val;
        } else {
          value -= pow_long(10,cursor);
          beep(10);
        }
      }
      led_display_number( value, cursor+1, ROW2, digits );
    }

    if( sw_on(3,DEFAULT_ON_TIME,DEFAULT_OFF_TIME) ) {
      if( cursor == (digits - 1) ) {
        cursor = 0;
      } else {
        cursor++;
      }
      beep(10);
      led_display_number( value, cursor+1, ROW2, digits );
    }
    if( sw_on(2,500,DEFAULT_OFF_TIME) ) {  // cancel & return
      value = init_val;
      beep(10);
      led_display_number( value, 0, ROW2, digits );
      delay_ms(1000);
      break;
    }
    if( sw_on(1,DEFAULT_ON_TIME,DEFAULT_OFF_TIME) ) {  // return
      beep(10);
      break;
    }
  }

  return value;
}


//----------------------------------------------------------------------------
// ロータリーエンコーダによる数値入力 long型 + カーソル位置
//
// init_val:初期値
// dp_val  :小数点位置
//
// return: 入力値x10 + カーソル位置
//----------------------------------------------------------------------------
unsigned long input_number_dp( unsigned long init_val, unsigned char dp_val )
{
  unsigned char turn;     // CW,CCW　ロータリーエンコーダ回転方向
  unsigned char cursor;   // カーソル位置
  unsigned long value;    // 入力した値を格納する変数
  unsigned char dig[8];
  unsigned long tmp_val;

  tmp_val = init_val;
  for( unsigned char index = 0 ; index < 8 ; index++ ) {
    dig[index] = tmp_val % 10;
    tmp_val /= 10;
  }
  cursor = dp_val;
  value = init_val;
  led_display_number( value, cursor+1, 2, 8 );
  while( TRUE ) {
    turn = rotary_encoder();
    if( turn != ENCODER_STOP ) {
      if( turn == CW ) {
        if( dig[cursor] != 9 ) {
          dig[cursor]++;
          beep(10);
        }
      } else {
        if( dig[cursor] != 0 ) {
          dig[cursor]--;
          beep(10);
        }
      }
      value = 0;
      for( unsigned char index = 0 ; index < 8 ; index++ ) {
        value *= 10;
        value += dig[7-index];
      }
      led_display_number( value, cursor+1, 2, 8 );
    }

    if( sw_on(3,DEFAULT_ON_TIME,DEFAULT_OFF_TIME) ) {
      if( cursor == 7 ) {
        cursor = 0;
      } else {
        cursor++;
      }
      beep(10);
      led_display_number( value, cursor+1, 2, 8 );
    }

    if( sw_on(2,500,DEFAULT_OFF_TIME) ) {  // cancel & return
      cursor = dp_val;
      value  = init_val;
      beep(10);
      led_display_number( value, cursor+1, 2, 8 );
      delay_ms(1000);
      break;
    }
    if( sw_on(1,DEFAULT_ON_TIME,DEFAULT_OFF_TIME) ) {  // return
      beep(10);
      break;
    }
  }

  return value*10 + cursor;
}

//----------------------------------------------------------------------------
// ロータリーエンコーダによる ON/OFF 入力
//
// init_val:初期値
//----------------------------------------------------------------------------
void display_onoff( unsigned char onoff )
{
  if( onoff == ENABLE ) {
    max7219_write_string( (unsigned char *)"     0n ", 1, 8, ROW2, 0 );
  } else {
    max7219_write_string( (unsigned char *)"     0FF", 1, 8, ROW2, 0 );
  }
}

unsigned char input_onoff( unsigned char onoff )
{
  unsigned char turn;     // CW,CCW　ロータリーエンコーダ回転方向
  unsigned char tmp_onoff;


  tmp_onoff = onoff;
  display_onoff( tmp_onoff );
  while( TRUE ) {
    turn = rotary_encoder();
    if( turn != ENCODER_STOP ) {
      if( turn == CW ) {
        if( tmp_onoff != ENABLE ) {
          tmp_onoff = ENABLE;
          beep(10);
        }
      } else {
        if( tmp_onoff != DISABLE ) {
          tmp_onoff = DISABLE;
          beep(10);
        }
      }
      display_onoff( tmp_onoff );
    }

    if( sw_on(1,DEFAULT_ON_TIME,DEFAULT_OFF_TIME) ) {  // return
      beep(10);
      break;
    }
  }

  return tmp_onoff;
}

//------------------------------------------------------------------------------
// input MAX7219 intensity
//
// dimmer: 0x01(dark) 0x0F(bright)
//------------------------------------------------------------------------------
unsigned char input_intensity(unsigned char initial_value)
{
  unsigned char turn;
  unsigned char intensity;

  max7219_clear();
  led_display_number( 88888888, 0, 2, 8 );
  intensity = initial_value;
  while( TRUE ) {
    turn = rotary_encoder();
    if( turn != ENCODER_STOP ) {
      if( turn == CW ) {
        if( intensity < 0x0F ) {
          intensity++;
        }
        beep(10);
      } else {
        if( 1 < intensity ) {
          intensity--;
        }
        beep(10);
      }
      max7219_intensity(intensity);
    }
    if( sw_on(2,DEFAULT_ON_TIME,DEFAULT_OFF_TIME) ) {
      break;
    }
  }

  return intensity;
}

//------------------------------------------------------------------------------
// program main
//------------------------------------------------------------------------------
int main(int argc, char** argv) {
  unsigned char turn;     // CW,CCW　ロータリーエンコーダ回転方向
  unsigned char intensity;
  unsigned long rpm;
  Union4bytes tmr_union;
  double tmr1_raw;
  double dbl_tmp1;
  double gear_ratio;
  unsigned long wheel_speed;
  unsigned long gear_ratio_long;
  unsigned char gear_ratio_dp;
  unsigned char relay_rpm_on;
  unsigned char relay_speed_on;
  unsigned long relay_on_upper_rpm;
  unsigned long relay_off_lower_rpm;
  unsigned long relay_on_upper_speed;
  unsigned long relay_off_lower_speed;
  unsigned long diameter;
  unsigned long diameter_min;
  unsigned long diameter_max;
  unsigned long diameter_step;
  unsigned char speed_display;
  unsigned int  display_update_time;
  unsigned int  zero_clear_time;
  unsigned long diameter_step_tmp;

  INTCON2bits.NOT_RBPU = 1;
  WDTCONbits.SWDTEN = 0; // WTD OFF

  led_hw_initialize();
  relay_hw_initialize();
  push_switch_hw_initialize();
  rotary_encoder_hw_initialize();
  buzzer_hw_initialize();

  Spi_initialize();
  max7219_initialize();
  delay_ms(1000);
  initialize_large_display();
  intensity = eeprom_read(EEPROM_ADDRESS_LED_INTENSITY);
  max7219_intensity( intensity );
  max7219_clear();

  gear_ratio_long       = read_long_eeprom(EEPROM_ADDRESS_GEAR_RATIO);
  gear_ratio_dp         = eeprom_read(EEPROM_ADDRESS_GEAR_RATIO_DP);
  gear_ratio            = (double)gear_ratio_long / (double)pow_long(10,gear_ratio_dp);
  relay_rpm_on          = eeprom_read(EEPROM_ADDRESS_RPM_RELAY_ENABLE);
  relay_on_upper_rpm    = read_long_eeprom(EEPROM_ADDRESS_RPM_TARGET_UPPER);
  relay_off_lower_rpm   = read_long_eeprom(EEPROM_ADDRESS_RPM_TARGET_LOWER);
  relay_speed_on        = eeprom_read(EEPROM_ADDRESS_SPEED_RELAY_ENABLE);
  relay_on_upper_speed  = read_long_eeprom(EEPROM_ADDRESS_SPEED_TARGET_UPPER);
  relay_off_lower_speed = read_long_eeprom(EEPROM_ADDRESS_SPEED_TARGET_LOWER);
  diameter              = read_long_eeprom(EEPROM_ADDRESS_DIAMETER_CURRENT);
  diameter_min          = read_long_eeprom(EEPROM_ADDRESS_DIAMETER_MIN);
  diameter_max          = read_long_eeprom(EEPROM_ADDRESS_DIAMETER_MAX);
  diameter_step         = read_long_eeprom(EEPROM_ADDRESS_DIAMETER_STEP);

  speed_display       = eeprom_read(EEPROM_ADDRESS_SPEED_ENABLE);
  zero_clear_time     = read_word_eeprom(EEPROM_ADDRESS_ZERO_CLEAR_TIME);
  display_update_time = read_word_eeprom(EEPROM_ADDRESS_UPDATE_TIME);
  display_update_count = (unsigned int)((double)display_update_time / TIMER1_OVERFLOW_TIME);

  sensor_hw_initialize();
  sensor1_interupt_enable();

  tmr2_initialize();
  tmr1_initialize();

  Interrupt_Priority_Level_Enable = TRUE;
  Interrupt_Low_Enable  = TRUE;
  Interrupt_High_Enable = TRUE;

  max7219_write_string( (unsigned char *)VERSION_STRING, 1, 8, ROW1, 6 );
  delay_ms(1000);

  // target = read_long_eeprom(0);
  // target_changed = TRUE;
  display_flag   = FALSE;
  start_flag     = FALSE;
  rpm = 0;

  while( TRUE ) {
    //------- input ----------------------------------
    if( sw_on(1,DEFAULT_ON_TIME,DEFAULT_OFF_TIME) ) {
      beep(10);
      LED1 = LED_ON;
      //----- Gear Ratio --------------------------
      max7219_clear();
      max7219_write_string( (unsigned char *)"rAtio", 1, 5, ROW1, 0 );
      gear_ratio_long = input_number_dp( gear_ratio_long, gear_ratio_dp );
      gear_ratio_dp = gear_ratio_long % 10;
      gear_ratio_long /=10;
      write_long_eeprom( EEPROM_ADDRESS_GEAR_RATIO, gear_ratio_long );
      eeprom_write( EEPROM_ADDRESS_GEAR_RATIO_DP  , gear_ratio_dp );
      gear_ratio = (double)gear_ratio_long / (double)pow_long(10,gear_ratio_dp);
      //----- Relay1 ON/OFF ------------------------
      max7219_clear();
      max7219_write_string( (unsigned char *)"rELAY rP", 1, 8, ROW1, 0 );
      relay_rpm_on = input_onoff( relay_rpm_on );
      eeprom_write( EEPROM_ADDRESS_RPM_RELAY_ENABLE, relay_rpm_on );
      //----- Target RPM ------------------------
      if( relay_rpm_on == ENABLE ) {
        //----- RELAY ON Upper RPM ---------------------
        max7219_clear();
        max7219_write_string( (unsigned char *)"UP", 1, 2, ROW1, 0 );
        relay_on_upper_rpm = input_number(relay_on_upper_rpm,1,99999999);
        write_long_eeprom( EEPROM_ADDRESS_RPM_TARGET_UPPER, relay_on_upper_rpm );
        //----- RELAY OFF Lower RPM ---------------------
        if( relay_off_lower_rpm >= relay_on_upper_rpm ) {
          relay_off_lower_rpm = relay_on_upper_rpm - 1;
        }
        max7219_clear();
        max7219_write_string( (unsigned char *)"Lou", 1, 3, ROW1, 0 );
        relay_off_lower_rpm = input_number(relay_off_lower_rpm,0,relay_on_upper_rpm-1);
        write_long_eeprom( EEPROM_ADDRESS_RPM_TARGET_LOWER, relay_off_lower_rpm );
      }
      //----- Speed display ON/OFF ------------------
      max7219_clear();
      max7219_write_string( (unsigned char *)"SPEEd", 1, 5, ROW1, 0 );
      speed_display = input_onoff( speed_display );
      eeprom_write( EEPROM_ADDRESS_SPEED_ENABLE, speed_display );      
      if( speed_display == ENABLE ) {
        //----- diameter Min --------------------------------
        max7219_clear();
        max7219_write_string( (unsigned char *)"dia Lou", 1, 7, ROW1, 0 );
        diameter_min = input_number( diameter_min, 1, 100000 );
        write_long_eeprom( EEPROM_ADDRESS_DIAMETER_MIN, diameter_min );
        //----- diameter Min --------------------------------
        max7219_clear();
        max7219_write_string( (unsigned char *)"dia UP", 1, 6, ROW1, 0 );
        diameter_max = input_number( diameter_max, diameter_min, 100000 );
        write_long_eeprom( EEPROM_ADDRESS_DIAMETER_MAX, diameter_max );
        //----- diameter Step -------------------------------
        max7219_clear();
        max7219_write_string( (unsigned char *)"dia StEP", 1, 8, ROW1, 0 );
        diameter_step = input_number( diameter_step, 1, 1000 );
        write_long_eeprom( EEPROM_ADDRESS_DIAMETER_STEP, diameter_step );
        //----- diameter current -----------------------------
        if( diameter < diameter_min ) {
          diameter = diameter_min;
        }
        if( diameter_max < diameter  ) {
          diameter = diameter_max;
        }
        write_long_eeprom( EEPROM_ADDRESS_DIAMETER_CURRENT, diameter );
        //----- Relay2 ON/OFF ------------------------
        max7219_clear();
        max7219_write_string( (unsigned char *)"rELAY SP", 1, 8, ROW1, 0 );
        relay_rpm_on = input_onoff( relay_speed_on );
        eeprom_write( EEPROM_ADDRESS_SPEED_RELAY_ENABLE, relay_speed_on );
        //----- Target Speed ------------------------
        if( relay_speed_on == ENABLE ) {
          //----- RELAY ON Upper Spped ---------------------
          max7219_clear();
          max7219_write_string( (unsigned char *)"UP", 1, 2, ROW1, 0 );
          relay_on_upper_speed = input_number(relay_on_upper_speed,1,99999999);
          write_long_eeprom( EEPROM_ADDRESS_SPEED_TARGET_UPPER, relay_on_upper_speed );
          //----- RELAY OFF Lower RPM ---------------------
          if( relay_off_lower_rpm >= relay_on_upper_rpm ) {
            relay_off_lower_rpm = relay_on_upper_rpm - 1;
          }
          max7219_clear();
          max7219_write_string( (unsigned char *)"Lou", 1, 3, ROW1, 0 );
          relay_off_lower_speed = input_number(relay_off_lower_speed,0,relay_on_upper_speed-1);
          write_long_eeprom( EEPROM_ADDRESS_RPM_TARGET_LOWER, relay_off_lower_speed );
        }
      }
      //----- display update time ---------------------
      max7219_clear();
      max7219_write_string( (unsigned char *)"UPdatE", 1, 6, ROW1, 0 );
      display_update_time = input_number(display_update_time,100,100000);
      write_word_eeprom(EEPROM_ADDRESS_UPDATE_TIME,display_update_time);
      display_update_count = (unsigned int)((double)display_update_time / TIMER1_OVERFLOW_TIME);
      max7219_clear();
      //----- zero clear time -------------------------
      max7219_clear();
      max7219_write_string( (unsigned char *)"0 CLEAr", 1, 7, ROW1, 0 );
      zero_clear_time = input_number(zero_clear_time,display_update_time,100000);
      write_word_eeprom(EEPROM_ADDRESS_ZERO_CLEAR_TIME,zero_clear_time);
      max7219_clear();
      LED1 = LED_OFF;
    }

    //------ input diameter ---------------------------
    if( speed_display == ENABLE ) {
      turn = rotary_encoder();
      if( turn != ENCODER_STOP ) {
        if( display_diameter_counter > 980 ) {
          diameter_step_tmp = diameter_step * 3;
        } else {
          diameter_step_tmp = diameter_step;
        }
        display_diameter_counter = 1000;
        if( turn == CW && diameter_max != diameter ) {
          if( diameter + diameter_step_tmp <= diameter_max ) {
            diameter += diameter_step_tmp;
          } else {
            diameter = diameter_max;
          }
          beep(10);
        }
        if( turn == CCW && diameter_min != diameter ) {
          if( diameter_min + diameter_step_tmp < diameter ) {
            diameter -= diameter_step_tmp;
          } else {
            diameter = diameter_min;
          }
          beep(10);
        }
        led_display_number( diameter, 0, ROW2, 8 );
        zero_display_counter = 0;
      }
    }

    //------- display ZERO ---------------------------
    if( zero_display_counter >= zero_clear_time ) {
      led_display_number( 0, 0, ROW1, 8 );
      if( speed_display == ENABLE ) {
        led_display_number( 0, 0, ROW2, 8 );
      }
      refresh_large_display();
      zero_display_counter = 0;
      rpm         = 0;
      wheel_speed = 0;
      start_flag = FALSE;
    }

    //------- display --------------------------------
    if( display_flag && display_diameter_counter == 0 ) {
      tmr_union.uchar[0] = TMR1L_value;
      tmr_union.uchar[1] = TMR1H_value;
      tmr_union.uint[1]  = TMR1_OVF_value;
      tmr1_raw = (double)tmr_union.ulong;

      dbl_tmp1 = CONST_MIN / tmr1_raw;
      rpm = (unsigned long)(dbl_tmp1 * (double)rotate_value * gear_ratio);

      led_display_number( rpm, 0, ROW1, 8 );
      refresh_large_display();

      if( speed_display == ENABLE ) {
        wheel_speed = (unsigned long)((double)rpm * (double)diameter * 3.1415926 / 1000);
        led_display_number( wheel_speed, 0, ROW2, 8 );
      }
      
      display_flag = FALSE;
    }

    //------ Relay1 ----------------------------------
    if( relay_on_upper_rpm <= rpm ) {
      LED4 = LED_ON;
      RELAY1 = RELAY_ON;
    }
    if( rpm <= relay_off_lower_rpm ) {
      LED4 = LED_OFF;
      RELAY1 = RELAY_OFF;
    }

    //------ Relay2 ----------------------------------
    if( relay_on_upper_speed <= wheel_speed ) {
      LED5 = LED_ON;
      RELAY2 = RELAY_ON;
    }
    if( wheel_speed <= relay_off_lower_speed ) {
      LED5 = LED_OFF;
      RELAY2 = RELAY_OFF;
    }

    //------ LED intensity ---------------------------
    if( sw_on(2,DEFAULT_ON_TIME,DEFAULT_OFF_TIME) ) {
      beep(10);
      LED2 = LED_ON;
      intensity = input_intensity( intensity );
      eeprom_write( EEPROM_ADDRESS_LED_INTENSITY, intensity );
      LED2 = LED_OFF;
    }

  }

  return (EXIT_SUCCESS);
}

